/*
 * Licensed Materials - Property of IBM
 *
 * com.ibm.etools.c
 *
 *  Copyright IBM Corporation 2004. All Rights Reserved. 
 *
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.etools.c.importer;

import java.util.Iterator;
import java.util.List;

import com.ibm.etools.c.CArray;
import com.ibm.etools.c.CClassifier;
import com.ibm.etools.c.CDatatype;
import com.ibm.etools.c.CDerivableType;
import com.ibm.etools.c.CEnumerator;
import com.ibm.etools.c.CField;
import com.ibm.etools.c.CFunction;
import com.ibm.etools.c.CParameter;
import com.ibm.etools.c.CPointer;
import com.ibm.etools.c.CStruct;
import com.ibm.etools.c.CStructured;
import com.ibm.etools.c.CTypedElement;
import com.ibm.etools.c.CTypedef;
import com.ibm.etools.c.CUnion;
import com.ibm.etools.c.datatypes.CBitField;
import com.ibm.etools.c.datatypes.CChar;
import com.ibm.etools.c.datatypes.CEnumeration;
import com.ibm.etools.c.datatypes.CInt;
import com.ibm.etools.c.datatypes.CLongDouble;
import com.ibm.etools.c.datatypes.CLongLong;
import com.ibm.etools.c.datatypes.CSignedChar;
import com.ibm.etools.c.datatypes.CUnsignedChar;
import com.ibm.etools.c.datatypes.CUnsignedInt;
import com.ibm.etools.c.datatypes.CUnsignedLong;
import com.ibm.etools.c.datatypes.CUnsignedLongLong;
import com.ibm.etools.c.datatypes.CUnsignedShort;
import com.ibm.etools.c.datatypes.CWchar;
import com.ibm.etools.tdlang.TDLangModelElement;

/**
 * The CDeclarationWriter class is used to translate C model objects back to
 * their original string representation.
 */
public class CDeclarationWriter {

    public static final String copyright = "Licensed Material - Property of IBM\n com.ibm.etools.c\n(C) Copyright IBM Corp. 2004 - All Rights Reserved.\nUS Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp."; //$NON-NLS-1$

    private static boolean _isOutputTD = false;

    private static CDeclarationWriter _instance = new CDeclarationWriter();

    /**
     * Constructors
     */
    private CDeclarationWriter() {
    }

    /**
     * Return the single instance
     */
    public static CDeclarationWriter instance() {
        return _instance;
    }

    /**
     * Set whether Type Descriptor output (offset and size) is generated.
     */
    public static void setOutputTD(boolean isOutputTD) {
        _isOutputTD = isOutputTD;
    }

    /**
     * Return the identifier of a C model element
     */
    public static String typeId(Object element) {

        String typeId = null;
        if (element instanceof CFunction) {
            typeId = "function";
        } else if (element instanceof CStruct) {
            typeId = "struct";
        } else if (element instanceof CUnion) {
            typeId = "union";
        } else if (element instanceof CTypedef) {
            typeId = "typedef";
        } else if (element instanceof CPointer) {
            typeId = "pointer";
        } else if (element instanceof CArray) {
            typeId = "array";
        } else if (element instanceof CField) {
            typeId = "field";
        } else if (element instanceof CParameter) {
            typeId = "parameter";
        } else if (element instanceof CEnumeration) {
            typeId = "enumeration";
        } else if (element instanceof CEnumerator) {
            typeId = "enumerator";
        } else if (element instanceof CDatatype) {
            typeId = "datatype";
        }

        return typeId;
    }

    /**
     * Return the short name of a C model object
     */
    public static String name(Object element) {

        String elementName = null;
        if (element instanceof CDatatype) {
            elementName = nameOfDatatype((CDatatype) element);
        } else if (element instanceof CPointer) {
            elementName = nameOfPointer((CPointer) element);
        } else if (element instanceof CArray) {
            elementName = nameOfArray((CArray) element);
        } else if (element instanceof CStruct) {
            elementName = "struct " + ((CStruct) element).getName();
        } else if (element instanceof CUnion) {
            elementName = "union " + ((CUnion) element).getName();
        } else if (element instanceof CEnumeration) {
            elementName = "enum " + ((CEnumeration) element).getName();
        } else if (element instanceof TDLangModelElement) {
            elementName = ((TDLangModelElement) element).getName();
        }

        return elementName;

    }

    /**
     * Return the name of a C primitive datatype as a string
     */
    public static String nameOfDatatype(CDatatype atype) {

        String typeName = null;
        if (atype instanceof CUnsignedChar) {
            typeName = "unsigned char";
        } else if (atype instanceof CSignedChar) {
            typeName = "signed char";
        } else if (atype instanceof CWchar) {
            typeName = "wchar_t";
        } else if (atype instanceof CChar) {
            typeName = "char";
        } else if (atype instanceof CUnsignedShort) {
            typeName = "unsigned short";
        } else if (atype instanceof CUnsignedLong) {
            typeName = "unsigned long";
        } else if (atype instanceof CUnsignedLongLong) {
            typeName = "unsigned longlong";
        } else if (atype instanceof CUnsignedInt) {
            typeName = "unsigned int";
        } else if (atype instanceof com.ibm.etools.c.datatypes.CShort) {
            typeName = "short";
        } else if (atype instanceof com.ibm.etools.c.datatypes.CLong) {
            typeName = "long";
        } else if (atype instanceof CLongLong) {
            typeName = "longlong";
        } else if (atype instanceof CInt) {
            typeName = "int";
        } else if (atype instanceof CEnumeration) {
            typeName = atype.getName();
        } else if (atype instanceof com.ibm.etools.c.datatypes.CFloat) {
            typeName = "float";
        } else if (atype instanceof com.ibm.etools.c.datatypes.CDouble) {
            typeName = "double";
        } else if (atype instanceof CLongDouble) {
            typeName = "long double";
        } else if (atype instanceof com.ibm.etools.c.datatypes.CVoid) {
            typeName = "void";
        } else if (atype instanceof CBitField) {
            CBitField abitfield = (CBitField) atype;
            typeName = nameOfDatatype(abitfield.getBaseType());
        }

        return typeName;

    }

    /**
     * Return the string representation of a pointer
     */
    public static String nameOfPointer(CPointer apointer) {

        String baseName = null;
        CDerivableType derivingType = apointer.getDerives();
        if (derivingType != null) {
            if (derivingType instanceof CFunction) {
                return nameOfFunctionPointer(apointer);
            } else {
                baseName = name(derivingType);
            }
        } else {
            baseName = "<unknown type>";
        }

        return baseName + "*";
    }

    /**
     * Return the string representation of an array
     */
    public static String nameOfArray(CArray array) {

        String baseName = null;
        CDerivableType derivingType = array.getDerives();
        if (derivingType != null) {
            baseName = name(derivingType);
        } else {
            baseName = "<unknown type>";
        }

        return baseName + "[" + array.getDimension() + "]";

    }

    /**
     * Return the full declaration string for an object list
     */
    public static String fullDeclarations(List objectList) {

        StringBuffer buffer = new StringBuffer();
        Iterator it = objectList.iterator();
        while (it.hasNext()) {
            Object element = it.next();
            buffer.append(declarationOf(element));
        }
        return buffer.toString();
    }

    /**
     * Return the string representation of all user defined types in an object
     * list
     */
    public static String declarations(List objectList) {

        StringBuffer buffer = new StringBuffer();
        Iterator it = objectList.iterator();
        while (it.hasNext()) {
            Object element = it.next();
            if (element instanceof CStructured || element instanceof CFunction || element instanceof CTypedef) {
                buffer.append(declarationOf(element));
                buffer.append("\n");
            }
        }
        return buffer.toString();
    }

    /**
     * Return the declaration string of an C model element It will call the
     * corresponding declaration method according to the element type.
     */
    public static String declarationOf(Object element) {

        String str = null;
        if (element instanceof CFunction) {
            str = functionDeclaration((CFunction) element);
        } else if (element instanceof CStruct) {
            str = structDeclaration((CStruct) element);
        } else if (element instanceof CUnion) {
            str = unionDeclaration((CUnion) element);
        } else if (element instanceof CTypedef) {
            str = typedefDeclaration((CTypedef) element);
        } else if (element instanceof CEnumeration) {
            str = enumDeclaration((CEnumeration) element);
        } else if (element instanceof CField) {
            str = fieldDeclaration((CField) element);
        } else if (element instanceof CParameter) {
            str = parameterDeclaration((CParameter) element);
        } else {
            str = null;
        }

        return str;
    }

    /**
     * Return the string representation of a function pointer
     */
    public static String nameOfFunctionPointer(CPointer funcPointer) {

        CDerivableType derivingType = funcPointer.getDerives();
        if (!(derivingType instanceof CFunction)) {
            CResource.writeLog(CDeclarationWriter.instance(), "nameOfFunctionPointer", CResource.getString("_ERROR_Invalid_Function_Pointer", funcPointer.getName()));
        }

        CFunction afunction = (CFunction) derivingType;

        String returnType = null;
        StringBuffer arguments = new StringBuffer();
        List parameters = afunction.getParameter();
        if (parameters != null) {
            Iterator it = parameters.iterator();
            int index = 0;
            while (it.hasNext()) {
                Object child = it.next();
                if (child instanceof CParameter) {
                    CParameter aparameter = (CParameter) child;
                    // Process return type
                    if (aparameter.getDirection().getName().equals(CIM.M_DIRECTION_RETURN)) {
                        returnType = typeName(aparameter);
                    }
                    // Process arguments
                    else if (aparameter.getDirection().getName().equals(CIM.M_DIRECTION_ARGUMENT)) {
                        if (index > 0) {
                            arguments.append(',');
                        }
                        arguments.append(parameterDeclaration(aparameter));
                        index = index + 1;
                    }
                }
            }
        }

        if (returnType == null || returnType.length() == 0) {
            return "(*)(" + arguments.toString() + ")";
        } else {
            return returnType + " " + "(*)(" + arguments.toString() + ")";
        }
    }

    /**
     * Return the declaration string of a function
     */
    public static String functionDeclaration(CFunction afunction) {

        String returnType = null;
        // Process function name
        String functionName = afunction.getName();
        if (functionName == null || functionName.length() == 0) {
            functionName = "<unnamed function>";
        }

        StringBuffer arguments = new StringBuffer();
        List parameters = afunction.getParameter();
        if (parameters != null) {
            Iterator it = parameters.iterator();
            int index = 0;
            while (it.hasNext()) {
                Object child = it.next();
                if (child instanceof CParameter) {
                    CParameter aparameter = (CParameter) child;
                    // Process return type
                    if (aparameter.getDirection().getName().equals(CIM.M_DIRECTION_RETURN)) {
                        returnType = typeName(aparameter);
                    }
                    // Process arguments
                    else if (aparameter.getDirection().getName().equals(CIM.M_DIRECTION_ARGUMENT)) {
                        if (index > 0) {
                            arguments.append(',');
                        }
                        arguments.append(parameterDeclaration(aparameter));
                        index = index + 1;
                    }
                }
            }
        }

        // Is it a varArg function?
        String varArgString = null;
        if (afunction.getIsVarArg().booleanValue()) {
            varArgString = ",...";
        } else {
            varArgString = "";
        }

        if (returnType == null || returnType.length() == 0) {
            return functionName + "(" + arguments.toString() + varArgString + ");\n";
        } else {
            return returnType + " " + functionName + "(" + arguments.toString() + varArgString + ");\n";
        }
    }

    /**
     * Return the declaration string of a structure
     */
    public static String structDeclaration(CStruct astruct) {
        return structDeclaration("", astruct);
    }

    /**
     * Return the declaration string of a structure for a given indent
     */
    public static String structDeclaration(String indent, CStruct astruct) {

        StringBuffer buffer = new StringBuffer();
        buffer.append(indent);
        buffer.append("struct ");

        // Process struct name
        String structName = astruct.getName();
        if (structName != null && structName.length() > 0) {
            buffer.append(structName);
        } else {
            buffer.append("<unnamed struct>");
        }
        buffer.append("\n");
        buffer.append(indent);
        buffer.append("{\n");

        // Process struct fields
        String localindent = indent + "   ";
        List fields = astruct.getContains();
        Iterator iterator = fields.iterator();
        while (iterator.hasNext()) {
            Object child = iterator.next();
            if (child instanceof CField) {
                buffer.append(localindent);
                buffer.append(fieldDeclaration((CField) child));
                buffer.append("\n");
            } else if (child instanceof CStruct) {
                buffer.append(structDeclaration(localindent, (CStruct) child));
            } else if (child instanceof CUnion) {
                buffer.append(unionDeclaration(localindent, (CUnion) child));
            } else if (child instanceof CEnumeration) {
                buffer.append(enumDeclaration(localindent, (CEnumeration) child));
            }
        }

        buffer.append(indent);
        buffer.append("};\n");
        return buffer.toString();
    }

    /**
     * Return the declaration string of a union
     */
    public static String unionDeclaration(CUnion aunion) {
        return unionDeclaration("", aunion);
    }

    /**
     * Return the declaration string of a union for a given indent
     */
    public static String unionDeclaration(String indent, CUnion aunion) {

        StringBuffer buffer = new StringBuffer();
        buffer.append(indent);
        buffer.append("union ");
        String unionName = aunion.getName();
        if (unionName != null && unionName.length() > 0) {
            buffer.append(unionName);
        } else {
            buffer.append("<unnamed union>");
        }
        buffer.append("\n");
        buffer.append(indent);
        buffer.append("{\n");

        String localindent = indent + "   ";
        List fields = aunion.getContains();
        Iterator iterator = fields.iterator();
        while (iterator.hasNext()) {
            Object child = iterator.next();
            if (child instanceof CField) {
                buffer.append(localindent);
                buffer.append(fieldDeclaration((CField) child));
                buffer.append("\n");
            } else if (child instanceof CStruct) {
                buffer.append(structDeclaration(localindent, (CStruct) child));
            } else if (child instanceof CUnion) {
                buffer.append(unionDeclaration(localindent, (CUnion) child));
            } else if (child instanceof CEnumeration) {
                buffer.append(enumDeclaration(localindent, (CEnumeration) child));
            }
        }

        buffer.append(indent);
        buffer.append("};\n");
        return buffer.toString();
    }

    /**
     * Return the declaration string of a typedef
     */
    public static String typedefDeclaration(CTypedef atypedef) {

        StringBuffer buffer = new StringBuffer();
        buffer.append("typedef ");
        CDerivableType derivingType = atypedef.getDerives();
        if (derivingType != null) {
            buffer.append(name(derivingType));
        } else {
            buffer.append("<unknown type>");
        }

        String typedefName = atypedef.getName();
        if (typedefName != null && typedefName.length() > 0) {
            // In the case the typedef is a function pointer
            if (isFunctionPointer(derivingType)) {
                String str = buffer.toString();
                int index = str.indexOf(CIM.P_FUNCTION_PTR);
                if (index != -1) {
                    buffer.insert(index + 2, typedefName);
                }
            } else {
                buffer.append(" " + typedefName);
            }
        } else {
            buffer.append(" " + "<unknown typedef>");
        }

        buffer.append(";\n");
        return buffer.toString();
    }

    /**
     * Return the declaration string of an enum
     */
    public static String enumDeclaration(CEnumeration anenum) {
        return enumDeclaration("", anenum);
    }

    /**
     * Return the declaration string of an enum for a given indent
     */
    public static String enumDeclaration(String indent, CEnumeration anenum) {

        StringBuffer buffer = new StringBuffer();
        buffer.append(indent);
        buffer.append("enum ");
        String enumName = anenum.getName();
        if (enumName == null || enumName.length() == 0) {
            enumName = "<unnamed enum>";
        }
        buffer.append(enumName);
        buffer.append(" {");

        // Process the enumerators
        List enumerators = anenum.getEnumerator();
        Iterator it = enumerators.iterator();

        int index = 0;
        while (it.hasNext()) {
            Object element = it.next();
            if (element instanceof CEnumerator) {

                if (index > 0) {
                    buffer.append(',');
                }

                CEnumerator enumerator = (CEnumerator) element;
                String enumeratorName = enumerator.getName();

                if (enumeratorName == null || enumeratorName.length() == 0) {
                    enumeratorName = "<unnamed enumerator>";
                }

                Integer enumeratorValue = enumerator.getValue();

                if (enumeratorValue == null) {
                    buffer.append(enumeratorName);
                } else {
                    buffer.append(enumeratorName + "=" + String.valueOf(enumeratorValue.intValue()));
                }
                index = index + 1;
            }
        }

        buffer.append("};\n");

        return buffer.toString();
    }

    /**
     * Return the declaration string of a typed element (e.g. a field or a
     * parameter)
     */
    public static String typedElementDeclaration(CTypedElement atypedElement) {

        String typeString = null;

        // Process the name
        String nameString = atypedElement.getName();
        if (nameString == null || nameString.length() == 0) {
            nameString = "";
        }

        CClassifier type = atypedElement.getType();
        if (type != null) {
            // Need to do special processing for arrays.
            // The array size is added after the name string rather than the
            // type string
            //
            if (type instanceof CArray) {
                CArray array = (CArray) type;
                CDerivableType derivingType = array;

                do {
                    array = (CArray) derivingType;
                    derivingType = array.getDerives();
                    Integer dimension = array.getDimension();
                    String dimensionString = null;

                    if (dimension == null || dimension.intValue() == 0) {
                        dimensionString = "<unknown dimension>";
                    } else {
                        dimensionString = String.valueOf(dimension.intValue());
                    }

                    nameString = nameString + "[" + dimensionString + "]";
                } while (derivingType instanceof CArray);

                typeString = name(derivingType);
            }

            // Need to do the same thing for bitfields
            else if (type instanceof CBitField) {
                CBitField abitfield = (CBitField) type;
                int size = abitfield.getSize().intValue();
                String sizeString = null;
                if (size < 0) {
                    sizeString = "<unknown size>";
                } else {
                    sizeString = String.valueOf(size);
                }
                nameString = nameString + ":" + sizeString;
                typeString = nameOfDatatype(abitfield.getBaseType());
            }
            // Function pointer needs special handling
            else if (isFunctionPointer(type)) {
                CPointer funcPointer = (CPointer) type;
                String funcPointerName = nameOfFunctionPointer(funcPointer);
                int index = funcPointerName.indexOf(')');
                StringBuffer buffer = new StringBuffer(funcPointerName);
                buffer.insert(index, nameString);
                return buffer.toString();
            } else {
                typeString = name(type);
            }

            if (typeString == null || typeString.length() == 0) {
                typeString = "<unknown type>";
            }
        } else {
            typeString = "<no type>";
        }

        String returnString = typeString;
        if (nameString.length() > 0) {
            returnString = typeString + " " + nameString;
        }

        return returnString;
    }

    /**
     * Return the declaration string of a struct/union field
     */
    public static String fieldDeclaration(CField afield) {
        String returnString = typedElementDeclaration(afield) + ";";
        if (_isOutputTD) {
            int offset = 0;
            int fieldSize = 0;
            try {
                offset = CTypeDescriptorHelper.offsetOf(afield);
                fieldSize = CTypeDescriptorHelper.sizeof(afield);
            } catch (Exception e) {
                CResource.writeLog(CDeclarationWriter.instance(), "fieldDeclaration", e);
            }

            returnString = returnString + "\t(" + offset + ", " + fieldSize + ")";
        }

        return returnString;
    }

    /**
     * Return the declaration string of a function parameter
     */
    public static String parameterDeclaration(CParameter aparameter) {
        return typedElementDeclaration(aparameter);
    }

    /**
     * Return the type name of a typed element
     */
    public static String typeName(CTypedElement atypedElement) {

        String typename = null;
        CClassifier type = atypedElement.getType();
        if (type != null) {
            typename = name(type);
        } else {
            typename = "<unknown type>";
        }

        return typename;
    }

    /**
     * Does this ecore type represent a function pointer?
     */
    public static boolean isFunctionPointer(Object type) {

        if (type instanceof CPointer) {
            CPointer apointer = (CPointer) type;
            return (apointer.getDerives() instanceof CFunction);
        } else {
            return false;
        }
    }

}